"""Transformations from 1D."""

__all__: tuple[str, ...] = ()


from plum import dispatch

import unxt as u
from unxt.quantity import is_any_quantity

import coordinax._src.vectors.custom_types as ct
from coordinax._src.vectors import d1, d2, d3
from coordinax._src.vectors.private_api import wrap_vconvert_impl_params

# =============================================================================
# CartesianPos1D


# -----------------------------------------------
# to 2D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.CartesianPos2D],
    from_vector: type[d1.CartesianPos1D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos1D -> CartesianPos2D.

    The `x` coordinate is converted to the `x` coordinate of the 2D system.
    The `y` coordinate is a keyword argument and defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.CartesianPos1D(x=u.Quantity(1.0, "km"))
    >>> x2 = cxv.vconvert(cxv.CartesianPos2D, x)
    >>> print(x2)
    <CartesianPos2D: (x, y) [km]
        [1. 0.]>

    >>> x3 = cxv.vconvert(cxv.CartesianPos3D, x, y=u.Quantity(14, "km"))
    >>> print(x3)
    <CartesianPos3D: (x, y, z) [km]
        [ 1. 14.  0.]>

    """
    x = p["x"]
    return {"x": x, "y": p.get("y", 0 * x)}, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.PolarPos],
    from_vector: type[d1.CartesianPos1D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos1D -> PolarPos.

    The `x` coordinate is converted to the radial coordinate `r`.
    The `phi` coordinate is a keyword argument and defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.CartesianPos1D(x=u.Quantity(1.0, "km"))
    >>> x2 = cxv.vconvert(cxv.PolarPos, x)
    >>> print(x2)
    <PolarPos: (r[km], phi[rad])
        [1. 0.]>

    >>> x3 = cxv.vconvert(cxv.PolarPos, x, phi=u.Quantity(14, "deg"))
    >>> print(x3)
    <PolarPos: (r[km], phi[deg])
        [ 1. 14.]>

    """
    r = p["x"]
    phi = p.get("phi", u.Quantity(0, "rad") if is_any_quantity(r) else 0)
    return {"r": r, "phi": phi}, (p.get("out_aux", {}) or {})


# -----------------------------------------------
# to 3D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d3.CartesianPos3D],
    from_vector: type[d1.CartesianPos1D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos1D -> CartesianPos3D.

    The `x` coordinate is converted to the `x` coordinate of the 3D system.
    The `y` and `z` coordinates are keyword arguments and default to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.CartesianPos1D(x=u.Quantity(1, "km"))
    >>> x2 = cxv.vconvert(cxv.CartesianPos3D, x)
    >>> print(x2)
    <CartesianPos3D: (x, y, z) [km]
        [1 0 0]>

    >>> x3 = cxv.vconvert(cxv.CartesianPos3D, x, y=u.Quantity(14, "km"))
    >>> print(x3)
    <CartesianPos3D: (x, y, z) [km]
        [ 1 14  0]>

    """
    x = p["x"]
    return {"x": x, "y": p.get("y", 0 * x), "z": p.get("z", 0 * x)}, (out_aux or {})


@dispatch.multi(
    (type[d3.SphericalPos], type[d1.CartesianPos1D], ct.ParamsDict),
    (type[d3.MathSphericalPos], type[d1.CartesianPos1D], ct.ParamsDict),
)
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d3.AbstractPos3D],
    from_vector: type[d1.CartesianPos1D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos1D -> SphericalPos | MathSphericalPos.

    The `x` coordinate is converted to the radial coordinate `r`.
    The `theta` and `phi` coordinates are keyword arguments and default to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    SphericalPos:

    >>> x = cxv.CartesianPos1D(x=u.Quantity(1, "km"))
    >>> x2 = cxv.vconvert(cxv.SphericalPos, x)
    >>> print(x2)
    <SphericalPos: (r[km], theta[rad], phi[rad])
        [1. 0. 0.]>

    >>> x3 = cxv.vconvert(cxv.SphericalPos, x, phi=u.Quantity(14, "deg"))
    >>> print(x3)
    <SphericalPos: (r[km], theta[rad], phi[deg])
        [ 1  0 14]>

    MathSphericalPos:
    Note that ``theta`` and ``phi`` have different meanings in this context.

    >>> x2 = cxv.vconvert(cxv.MathSphericalPos, x)
    >>> print(x2)
    <MathSphericalPos: (r[km], theta[rad], phi[rad])
        [1. 0. 0.]>

    >>> x3 = cxv.vconvert(cxv.MathSphericalPos, x, phi=u.Quantity(14, "deg"))
    >>> print(x3)
    <MathSphericalPos: (r[km], theta[rad], phi[deg])
        [ 1.  0. 14.]>

    """
    r = p["x"]
    theta = p.get("theta", u.Quantity(0, "rad") if is_any_quantity(r) else 0)
    phi = p.get("phi", u.Quantity(0, "rad") if is_any_quantity(r) else 0)
    return {"r": r, "theta": theta, "phi": phi}, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d3.CylindricalPos],
    from_vector: type[d1.CartesianPos1D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """RadialPos -> CartesianPos2D.

    The `x` coordinate is converted to the radial coordinate `rho`.
    The `phi` and `z` coordinates are keyword arguments and default to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.CartesianPos1D(x=u.Quantity(1, "km"))
    >>> x2 = cxv.vconvert(cxv.CylindricalPos, x)
    >>> print(x2)
    <CylindricalPos: (rho[km], phi[rad], z[km])
        [1. 0. 0.]>

    >>> x3 = cxv.vconvert(cxv.CylindricalPos, x, phi=u.Quantity(14, "deg"))
    >>> print(x3)
    <CylindricalPos: (rho[km], phi[deg], z[km])
        [ 1 14  0]>

    """
    rho = p["x"]
    phi = p.get("phi", u.Quantity(0, "rad") if is_any_quantity(rho) else 0)
    z = p.get("z", 0 * rho)
    return {"rho": rho, "phi": phi, "z": z}, (out_aux or {})


# =============================================================================
# RadialPos

# -----------------------------------------------
# 2D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.CartesianPos2D],
    from_vector: type[d1.RadialPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """RadialPos -> CartesianPos2D.

    The `r` coordinate is converted to the cartesian coordinate `x`.
    The `y` coordinate is a keyword argument and defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.RadialPos(r=u.Quantity(1, "km"))
    >>> x2 = cxv.vconvert(cxv.CartesianPos2D, x)
    >>> print(x2)
    <CartesianPos2D: (x, y) [km]
        [1 0]>

    >>> x3 = cxv.vconvert(cxv.CartesianPos2D, x, y=u.Quantity(14, "km"))
    >>> print(x3)
    <CartesianPos2D: (x, y) [km]
        [ 1 14]>

    """
    r = p["r"]
    return {"x": r, "y": p.get("y", 0 * r)}, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.PolarPos],
    from_vector: type[d1.RadialPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """RadialPos -> PolarPos.

    The `r` coordinate is converted to the radial coordinate `r`.
    The `phi` coordinate is a keyword argument and defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.RadialPos(r=u.Quantity(1, "km"))
    >>> x2 = cxv.vconvert(cxv.PolarPos, x)
    >>> print(x2)
    <PolarPos: (r[km], phi[rad])
        [1. 0.]>

    >>> x3 = cxv.vconvert(cxv.PolarPos, x, phi=u.Quantity(14, "deg"))
    >>> print(x3)
    <PolarPos: (r[km], phi[deg])
        [ 1 14]>

    """
    newp = {
        "r": p["r"],
        "phi": p.get("phi", u.Quantity(0, "rad") if is_any_quantity(p["r"]) else 0),
    }
    return newp, (out_aux or {})


# -----------------------------------------------
# 3D


# TODO: I don't like this rule. C3->R1 takes the norm. This places it on the x
# axis. What about giving theta, phi instead?
@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d3.CartesianPos3D],
    from_vector: type[d1.RadialPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """RadialPos -> CartesianPos3D.

    The `r` coordinate is converted to the `x` coordinate of the 3D system.
    The `y` and `z` coordinates are keyword arguments and default to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.RadialPos(r=u.Quantity(1.0, "km"))
    >>> y = cxv.vconvert(cxv.CartesianPos3D, x)
    >>> print(y)
    <CartesianPos3D: (x, y, z) [km]
        [1. 0. 0.]>

    >>> y = cxv.vconvert(cxv.CartesianPos3D, x, y=u.Quantity(14, "km"))
    >>> print(y)
    <CartesianPos3D: (x, y, z) [km]
        [ 1. 14.  0.]>

    """
    r = p["r"]
    newp = {"x": r, "y": p.get("y", 0 * r), "z": p.get("z", 0 * r)}
    return newp, (out_aux or {})


@dispatch.multi(
    (type[d3.SphericalPos], type[d1.RadialPos], ct.ParamsDict),
    (type[d3.MathSphericalPos], type[d1.RadialPos], ct.ParamsDict),
)
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d3.AbstractPos3D],
    from_vector: type[d1.RadialPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """RadialPos -> SphericalPos | MathSphericalPos.

    The `r` coordinate is converted to the radial coordinate `r`.
    The `theta` and `phi` coordinates are keyword arguments and default to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.RadialPos(r=u.Quantity(1, "km"))

    SphericalPos:

    >>> y = cxv.vconvert(cxv.SphericalPos, x)
    >>> print(y)
    <SphericalPos: (r[km], theta[rad], phi[rad])
        [1. 0. 0.]>

    >>> y = cxv.vconvert(cxv.SphericalPos, x, phi=u.Quantity(14, "deg"))
    >>> print(y)
    <SphericalPos: (r[km], theta[rad], phi[deg])
        [ 1  0 14]>

    MathSphericalPos:

    >>> y = cxv.vconvert(cxv.MathSphericalPos, x)
    >>> print(y)
    <MathSphericalPos: (r[km], theta[rad], phi[rad])
        [1. 0. 0.]>

    >>> y = cxv.vconvert(cxv.MathSphericalPos, x, phi=u.Quantity(14, "deg"))
    >>> print(y)
    <MathSphericalPos: (r[km], theta[rad], phi[deg])
        [ 1.  0. 14.]>

    """
    r = p["r"]
    newp = {
        "r": r,
        "theta": p.get("theta", u.Quantity(0, "rad") if is_any_quantity(r) else 0),
        "phi": p.get("phi", u.Quantity(0, "rad") if is_any_quantity(r) else 0),
    }
    return newp, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d3.CylindricalPos],
    from_vector: type[d1.RadialPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """RadialPos -> CylindricalPos.

    The `r` coordinate is converted to the radial coordinate `rho`.
    The `phi` and `z` coordinates are keyword arguments and default to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.RadialPos(r=u.Quantity(1.0, "km"))
    >>> y = cxv.vconvert(cxv.CylindricalPos, x)
    >>> print(y)
    <CylindricalPos: (rho[km], phi[rad], z[km])
        [1. 0. 0.]>

    >>> y = cxv.vconvert(cxv.CylindricalPos, x, phi=u.Quantity(14, "deg"))
    >>> print(y)
    <CylindricalPos: (rho[km], phi[deg], z[km])
        [ 1. 14.  0.]>

    """
    r = p["r"]
    newp = {
        "rho": r,
        "phi": p.get("phi", u.Quantity(0, "rad") if is_any_quantity(r) else 0),
        "z": p.get("z", 0 * r),
    }
    return newp, (out_aux or {})



"""`coordinax.angle` module."""

__all__ = ("AbstractAngle", "Angle", "wrap_to", "Parallax")

from .setup_package import install_import_hook

with install_import_hook("coordinax.angle"):
    from unxt.quantity import AbstractAngle, Angle, wrap_to

    from ._src.distances import Parallax


del install_import_hook



"""Tests for `coordinax.KinematicSpace`."""

from textwrap import dedent

import coordinax as cx


def test_repr():
    """Test the repr of KinematicSpace."""
    # A simple KinematicSpace from a position
    space = cx.KinematicSpace.from_(cx.CartesianPos3D.from_([1, 2, 3], "kpc"))

    # Expected repr
    exp = """
    KinematicSpace({
      'length':
      CartesianPos3D(
        x=Quantity(1, unit='kpc'), y=Quantity(2, unit='kpc'), z=Quantity(3, unit='kpc')
      )
    })
    """[1:-1]  # remove first and last newline
    exp = dedent(exp).strip()  # dedent & strip leading/trailing whitespace

    assert repr(space) == exp



"""Representation of coordinates in different systems."""

__all__ = ("AbstractAcc1D", "AbstractPos1D", "AbstractVel1D")


from coordinax._src.vectors.base_acc import AbstractAcc
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel


class AbstractPos1D(AbstractPos):
    """Abstract representation of 1D coordinates in different systems."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianPos1D._dimensionality()
        1

        """
        return 1


class AbstractVel1D(AbstractVel):
    """Abstract representation of 1D differentials in different systems."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianVel1D._dimensionality()
        1

        """
        return 1


class AbstractAcc1D(AbstractAcc):
    """Abstract representation of 1D acceleration in different systems."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianAcc1D._dimensionality()
        1

        """
        return 1



"""Coordinates."""

__all__ = ("AbstractCoordinate", "Coordinate")


from typing import Any, ClassVar, Literal, assert_never, cast, final
from typing_extensions import override

import equinox as eqx
import jax
import wadler_lindig as wl
from plum import dispatch

import unxt as u
from dataclassish import field_items, replace
from dataclassish.converters import Unless

from .base import AbstractReferenceFrame
from .xfm import TransformedReferenceFrame
from coordinax._src.operators import Identity
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.collection.core import KinematicSpace


# TODO: parametrize by the vector type(s), when Space is parametrized,
# and the frame type(s)
class AbstractCoordinate(AbstractVector):
    """Coordinates are vectors in a reference frame.

    See Also
    --------
    `coordinax.Coordinate` for a concrete implementation.

    """

    #: The data of the coordinate. This is a `coordinax.KinematicSpace` object,
    #: which is a collection of vectors.
    data: eqx.AbstractVar[KinematicSpace]  # TODO: KinematicSpace[PosT] -- plum#212

    #: The reference frame of the coordinate as a
    #: `coordinax.frames.AbstractReferenceFrame` object.
    frame: eqx.AbstractVar[AbstractReferenceFrame]

    # ===============================================================
    # Coordinate API

    def to_frame(
        self, toframe: AbstractReferenceFrame, /, t: u.Quantity | None = None
    ) -> "AbstractCoordinate":
        """Transform the coordinate to a specified frame.

        Examples
        --------
        >>> import coordinax as cx

        >>> cicrs = cx.Coordinate(cx.CartesianPos3D.from_([1, 2, 3], "kpc"),
        ...                       cx.frames.ICRS())

        >>> cicrs.to_frame(cx.frames.ICRS()) is cicrs
        True

        >>> cgcf = cicrs.to_frame(cx.frames.Galactocentric())
        >>> cgcf
        Coordinate(
            KinematicSpace({ 'length': CartesianPos3D( ... ) }),
            frame=Galactocentric( ... )
        )

        """
        op = self.frame.transform_op(toframe)

        # Special case for identity operations
        if isinstance(op, Identity):
            return self

        # Otherwise, apply the transformation and return a new coordinate
        new_data = op(self.data) if t is None else op(t, self.data)[1]
        out = self.__class__.from_(new_data, toframe)
        return cast("AbstractCoordinate", out)

    # ===============================================================
    # Quax API

    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array."""
        return self.data.aval()

    # ===============================================================
    # Plum API

    __faithful__: ClassVar = True

    # ===============================================================
    # Wadler-Lindig API

    def __pdoc__(
        self,
        *,
        include_data_name: Literal["named", "vector", "map"] = "named",
        **kwargs: Any,
    ) -> wl.AbstractDoc:
        """Return the Wadler-Lindig representation.

        Parameters
        ----------
        include_data_name : {'named', 'vector', 'map'}, optional
            If `named`, include the name of the data field.
            If `vector`, only include the data itself as a `KinematicSpace`.
            If `map`, only include the data as dict-like representation.
        **kwargs : Any, optional
            Additional keyword arguments to pass to `wl.pdoc`.
            'vector_form' is one option.

        Examples
        --------
        >>> import coordinax as cx
        >>> import wadler_lindig as wl

        >>> frame = cx.frames.ICRS()
        >>> data = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
        >>> coord = cx.Coordinate(data, frame)

        >>> wl.pprint(coord, include_data_name="named")
        Coordinate(
            data=KinematicSpace({ 'length': CartesianPos3D( ... ) }),
            frame=ICRS()
        )

        >>> wl.pprint(coord, include_data_name="vector")
        Coordinate(
            KinematicSpace({ 'length': CartesianPos3D( ... ) }),
            frame=ICRS()
        )

        >>> wl.pprint(coord, include_data_name="map")
        Coordinate(
            { 'length': CartesianPos3D( ... ) },
            frame=ICRS()
        )

        >>> print(repr(coord))
        Coordinate(
            KinematicSpace({ 'length': CartesianPos3D( ... ) }),
            frame=ICRS()
        )

        >>> print(str(coord))
        Coordinate(
            {
                'length':
                <CartesianPos3D: (x, y, z) [kpc]
                    [1 2 3]>
            },
            frame=ICRS()
        )

        """
        match include_data_name:
            case "named":
                docs = wl.named_objs(tuple(field_items(self)), **kwargs)
            case "vector":
                docs = [
                    wl.pdoc(self.data, **kwargs),
                    *wl.named_objs(tuple(field_items(self))[1:], **kwargs),
                ]
            case "map":
                docs = [
                    wl.pdoc(self.data._data, **kwargs),
                    *wl.named_objs(tuple(field_items(self))[1:], **kwargs),
                ]
            case _:
                assert_never()

        return wl.bracketed(
            begin=wl.TextDoc(f"{self.__class__.__name__}("),
            docs=docs,
            sep=wl.comma,
            end=wl.TextDoc(")"),
            indent=kwargs.get("indent", 4),
        )

    # ===============================================================
    # Python API

    def __repr__(self) -> str:
        """Return string representation.

        Examples
        --------
        >>> import coordinax as cx
        >>> coord = cx.Coordinate(cx.CartesianPos3D.from_([1, 2, 3], "kpc"),
        ...                       cx.frames.ICRS())
        >>> print(repr(coord))
        Coordinate(
            KinematicSpace({ 'length': CartesianPos3D( ... ) }),
            frame=ICRS()
        )

        """
        return wl.pformat(self, width=88, include_data_name="vector", vector_form=False)

    def __str__(self) -> str:
        """Return string representation.

        Examples
        --------
        >>> import coordinax as cx
        >>> coord = cx.Coordinate(cx.CartesianPos3D.from_([1, 2, 3], "kpc"),
        ...                       cx.frames.ICRS())
        >>> print(coord)
        Coordinate(
            {
            'length': <CartesianPos3D: (x, y, z) [kpc]
                [1 2 3]>
            },
            frame=ICRS()
        )

        """
        return wl.pformat(self, width=88, include_data_name="map", vector_form=True)

    # ===============================================================
    # IPython API

    _repr_latex_ = lambda self: wl.pformat(self)  # noqa: E731  # TODO: implement this


##############################################################################


@final
class Coordinate(AbstractCoordinate):
    """Coordinates are vectors in a reference frame.

    Examples
    --------
    >>> import coordinax as cx

    >>> coord = cx.Coordinate(cx.CartesianPos3D.from_([1, 2, 3], "kpc"),
    ...                       cx.frames.ICRS())
    >>> coord
    Coordinate(
        KinematicSpace({ 'length': CartesianPos3D( ... ) }),
        frame=ICRS()
    )

    Alternative Construction:

    >>> frame = cx.frames.ICRS()
    >>> data = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> cx.Coordinate.from_({"data": data, "frame": frame})
    Coordinate(
        KinematicSpace({ 'length': CartesianPos3D( ... ) }),
        frame=ICRS()
    )

    Changing Representation:

    >>> frame = cx.frames.ICRS()
    >>> data = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> coord = cx.Coordinate(data, frame)

    >>> coord.vconvert(cx.SphericalPos)
    Coordinate(
        KinematicSpace({ 'length': SphericalPos( ... ) }),
        frame=ICRS()
    )

    Showing Frame Transformation:

    >>> space = cx.KinematicSpace(
    ...     length=cx.CartesianPos3D.from_([1.0, 0, 0], "pc"),
    ...     speed=cx.CartesianVel3D.from_([1.0, 0, 0], "km/s"))

    >>> w=cx.Coordinate(
    ...     data=space,
    ...     frame=cx.frames.TransformedReferenceFrame(
    ...         cx.frames.Galactocentric(),
    ...         cx.ops.GalileanSpatialTranslation.from_([20, 0, 0], "kpc"),
    ...     ),
    ... )

    >>> w.to_frame(cx.frames.ICRS())
    Coordinate(
        KinematicSpace({
            'length': CartesianPos3D(...), 'speed': CartesianVel3D(...) }),
        frame=ICRS()
    )

    >>> w.to_frame(cx.frames.ICRS()).data["length"]
    CartesianPos3D(
      x=Quantity(-1587.6683, unit='pc'),
      y=Quantity(-24573.762, unit='pc'),
      z=Quantity(-13583.504, unit='pc')
    )

    """

    # The data of the coordinate. This is a `coordinax.KinematicSpace` object,
    # which is a collection of vectors. This can be constructed from a space
    # object, or any input that can construct a `coordinax.KinematicSpace` via
    # `coordinax.KinematicSpace.from_`.
    data: KinematicSpace = eqx.field(
        converter=Unless(KinematicSpace, KinematicSpace.from_)
    )

    #: The reference frame of the coordinate as a
    #: `coordinax.frames.AbstractReferenceFrame` object. This can be
    #: from a reference frame object, or any input that can construct a
    #: `coordinax.frames.TransformedReferenceFrame` via
    #: `coordinax.frames.AbstractReferenceFrame.from_`.
    frame: AbstractReferenceFrame = eqx.field(
        converter=Unless(AbstractReferenceFrame, TransformedReferenceFrame.from_)
    )

    # ===============================================================
    # Vector API

    @override
    def _dimensionality(self) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> w = cx.Coordinate(cx.CartesianPos3D.from_([1, 2, 3], "kpc"),
        ...                   cx.frames.ICRS())
        >>> try: w._dimensionality()
        ... except NotImplementedError as e: print("not implemented")
        not implemented

        """
        # TODO: KinematicSpace is currently not implemented.
        return self.data._dimensionality()

    @dispatch
    def __getitem__(self: "Coordinate", index: Any) -> "Coordinate":
        """Return Coordinate, with indexing applied to the data.

        Examples
        --------
        >>> import coordinax as cx

        >>> data = cx.CartesianPos3D.from_([[1, 2, 3], [4, 5, 6]], "kpc")
        >>> w = cx.Coordinate.from_(data, cx.frames.ICRS())

        >>> print(w[0].data["length"])
        <CartesianPos3D: (x, y, z) [kpc]
            [1 2 3]>

        """
        return replace(self, data=self.data[index])

    @dispatch
    def __getitem__(self: "Coordinate", index: str) -> AbstractVector:
        """Return the data of the coordinate.

        Examples
        --------
        >>> import coordinax as cx

        >>> data = cx.CartesianPos3D.from_([[1, 2, 3], [4, 5, 6]], "kpc")
        >>> w = cx.Coordinate.from_(data, cx.frames.ICRS())

        >>> print(w["length"])
        <CartesianPos3D: (x, y, z) [kpc]
            [[1 2 3]
             [4 5 6]]>

        """
        return self.data[index]
